"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Agile = void 0;
const internal_1 = require("./internal");
class Agile {
    constructor(config = {}) {
        this.config = config;
        //=========================================================================================================
        // API
        //=========================================================================================================
        /**
         * Create Agile API
         * @param config Object
         * @param config.options Object - Typescript default: RequestInit (headers, credentials, mode, etc...)
         * @param config.baseURL String - Url to prepend to endpoints (without trailing slash)
         * @param config.timeout Number - Time to wait for request before throwing error
         */
        // public API = (config: apiConfig) => new API(config);
        //=========================================================================================================
        // Storage
        //=========================================================================================================
        /**
         * Create Agile Storage
         */
        this.Storage = (config) => new internal_1.Storage(this, config);
        //=========================================================================================================
        // State
        //=========================================================================================================
        /**
         * Create Agile State
         * @param initialState Any - the value to initialize a State instance with
         * @key State key/name which identifies the state
         */
        this.State = (initialState, key) => new internal_1.State(this, initialState, key);
        //=========================================================================================================
        // Collection
        //=========================================================================================================
        /**
         * Create Agile Collection
         * @param config object | function returning object
         * @param config.primaryKey string - The primary key for the collection.
         * @param config.groups object - Define groups for this collection.
         */
        this.Collection = (config) => new internal_1.Collection(this, config);
        //=========================================================================================================
        // Computed
        //=========================================================================================================
        /**
         * Create a Agile computed function
         * @param deps Array - An array of state items to depend on
         * @param computeFunction Function - A function where the return value is the state, ran every time a dep changes
         */
        this.Computed = (computeFunction, deps) => new internal_1.Computed(this, computeFunction, deps);
        //=========================================================================================================
        // Event
        //=========================================================================================================
        /**
         * Create a Pulse Event
         */
        this.Event = (config) => new internal_1.Event(this, config);
        this.integrations = new internal_1.Integrations(this);
        this.subController = new internal_1.SubController(this);
        this.runtime = new internal_1.Runtime(this);
        this.storage = new internal_1.Storage(this, config.storageConfig || {});
        // Bind Frameworks to Agile
        this.integrations.bind();
        // Creates a global agile instance..
        this.globalBind();
    }
    //=========================================================================================================
    // Use
    //=========================================================================================================
    /**
     * Use custom Integration
     */
    use(integration) {
        this.integrations.integrate(integration);
        return this;
    }
    //=========================================================================================================
    // Set Storage
    //=========================================================================================================
    /**
     * Configures the Agile Storage
     * @param storageConfig
     */
    setStorage(storageConfig) {
        // Get States which are already saved into a storage
        const persistedStates = this.storage.persistedStates;
        // Define new Storage
        this.storage = new internal_1.Storage(this, storageConfig);
        this.storage.persistedStates = persistedStates;
        // Save all already saved states into the new Storage
        this.storage.persistedStates.forEach(state => state.persist(state.key));
        // Save all already saved collections into the new Storage
        this.storage.persistedCollections.forEach(collection => collection.persist(collection.key));
    }
    //=========================================================================================================
    // Global Bind
    //=========================================================================================================
    /**
     * @internal
     * Creates a global reference to the first pulse instance created this runtime
     */
    globalBind() {
        try {
            // https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/globalThis
            // @ts-ignore
            if (!globalThis.__agile)
                globalThis.__agile = this;
        }
        catch (error) {
            // fail silently
        }
    }
}
exports.Agile = Agile;
Agile.initialIntegrations = [];
