"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Group = void 0;
const internal_1 = require("../internal");
const perstist_1 = require("./perstist");
class Group extends internal_1.State {
    constructor(agileInstance, collection, initialItems, config) {
        super(agileInstance, initialItems || [], config === null || config === void 0 ? void 0 : config.key);
        this._output = []; // Output of the group (Note: _value are only the keys of the collection items)
        this._states = []; // States of the Group
        this.notFoundPrimaryKeys = []; // Contains all key which can't be found in the collection
        this.collection = () => collection;
        // Set build() to state sideEffect
        this.sideEffects = () => this.build();
        // Set type of State to array because a group is an array of collection item keys
        this.type(Array);
        // Initial Build
        this.build();
    }
    get output() {
        // Add state(group) to foundState (for auto tracking used states in computed functions)
        if (this.agileInstance().runtime.trackState)
            this.agileInstance().runtime.foundStates.add(this);
        return this._output;
    }
    get states() {
        // Add state(group) to foundState (for auto tracking used states in computed functions)
        if (this.agileInstance().runtime.trackState)
            this.agileInstance().runtime.foundStates.add(this);
        return this._states.map(state => state());
    }
    //=========================================================================================================
    // Has
    //=========================================================================================================
    /**
     * Checks if the group contains the primaryKey
     */
    has(primaryKey) {
        return this.value.findIndex(key => key === primaryKey) !== -1;
    }
    //=========================================================================================================
    // Size
    //=========================================================================================================
    /**
     * Returns the size of the group
     */
    get size() {
        return this.value.length;
    }
    //=========================================================================================================
    // Remove
    //=========================================================================================================
    /**
     * Removes a item at primaryKey from the group
     */
    remove(itemKeys, options = {}) {
        const _itemKeys = internal_1.normalizeArray(itemKeys);
        const notExistingCollectionItems = [];
        // Merge default values into options
        options = internal_1.defineConfig(options, {
            background: false
        });
        _itemKeys.forEach(itemKey => {
            // If item doesn't exist in collection add it to notExistingItems
            if (!this.collection().findById(itemKey))
                notExistingCollectionItems.push(itemKey);
            // Check if primaryKey exists in group if not, return
            if (this.value.findIndex(key => key === itemKey) === -1) {
                console.error(`Agile: Couldn't find primaryKey '${itemKey}' in group`, this);
                return;
            }
            // Remove primaryKey from nextState
            this.nextState = this.nextState.filter((i) => i !== itemKey);
            // Storage
            if (this.key)
                perstist_1.updateGroup(this.key, this.collection());
        });
        // If all items don't exist in collection.. set background to true because the output won't change -> no rerender necessary
        if (notExistingCollectionItems.length >= _itemKeys.length)
            options.background = true;
        // Set State to nextState
        this.ingest(options);
        return this;
    }
    //=========================================================================================================
    // Add
    //=========================================================================================================
    /**
     * Adds a key to a group
     */
    add(itemKeys, options = {}) {
        const _itemKeys = internal_1.normalizeArray(itemKeys);
        const notExistingCollectionItems = [];
        let newNextState = [...this.nextState]; // Had to create copy array otherwise also 'this.value' would change.. by changing 'this.nextState' directly.
        // Merge default values into options
        options = internal_1.defineConfig(options, {
            method: 'push',
            overwrite: false,
            background: false
        });
        _itemKeys.forEach(itemKey => {
            // Check if item already exists in group
            const existsInGroup = newNextState.findIndex(key => key === itemKey) !== -1;
            // If item doesn't exist in collection add it to notExistingItems
            if (!this.collection().findById(itemKey))
                notExistingCollectionItems.push(itemKey);
            // Removes temporary key from group to overwrite it properly
            if (options.overwrite)
                newNextState = newNextState.filter((i) => i !== itemKey);
            // If we do not want to overwrite and key already exists in group, exit
            else if (existsInGroup)
                return;
            // Push or unshift into state
            newNextState[options.method || 'push'](itemKey);
            // Storage
            if (this.key)
                perstist_1.updateGroup(this.key, this.collection());
        });
        // If all items don't exist in collection.. set background to true because the output won't change -> no rerender necessary
        if (notExistingCollectionItems.length >= _itemKeys.length)
            options.background = true;
        // Set nextState to newNextState
        this.nextState = newNextState;
        // Set State to nextState
        this.ingest({ background: options.background });
        return this;
    }
    //=========================================================================================================
    // Build
    //=========================================================================================================
    /**
     * @internal
     * Will build the group -> it will set the output to the collection values
     */
    build() {
        this.notFoundPrimaryKeys = [];
        // Check if _value is an array if not something went wrong because a group is always an array
        if (!Array.isArray(this._value)) {
            console.error("Agile: A group state has to be an array!");
            return;
        }
        // Map though group _value (collectionKey array) and get their state from collection
        const finalStates = this._value
            .map((primaryKey) => {
            // Get collection data at the primaryKey position
            let data = this.collection().data[primaryKey];
            // If no data found add this key to missing PrimaryKeys
            if (!data) {
                this.notFoundPrimaryKeys.push(primaryKey);
                return;
            }
            return data;
        }).filter(item => item !== undefined);
        // Map though found States and return their publicValue
        const finalOutput = finalStates
            .map((state) => {
            // @ts-ignore
            return state.getPublicValue();
        });
        // Log not found primaryKeys
        if (this.notFoundPrimaryKeys.length > 0 && this.agileInstance().config.logJobs)
            console.warn(`Agile: Couldn't find states with the primary keys in group '${this.key}'`, this.notFoundPrimaryKeys);
        // @ts-ignore
        this._states = finalStates.map(state => (() => state));
        this._output = finalOutput;
    }
}
exports.Group = Group;
