"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Collection = void 0;
const internal_1 = require("../internal");
const perstist_1 = require("./perstist");
class Collection {
    constructor(agileInstance, config = {}) {
        this.size = 0; // The amount of data items stored inside this collection
        this.data = {}; // Collection data is stored here
        this.isPersistCollection = false; // Is saved in storage
        this.groups = {};
        this.selectors = {};
        this.agileInstance = () => agileInstance;
        // If collection config is a function, execute and assign to config
        if (typeof config === 'function')
            config = config(this);
        // Assign defaults to config
        this.config = internal_1.defineConfig(config, {
            primaryKey: 'id',
            groups: {},
            selectors: {},
            defaultGroupKey: 'default'
        });
        // Set Key
        this._key = this.config.key;
        // Init Groups
        this.initSubInstances('groups');
        // Init Selectors
        this.initSubInstances('selectors');
    }
    set key(value) {
        this._key = value;
    }
    get key() {
        return this._key;
    }
    //=========================================================================================================
    // Init SubInstances
    //=========================================================================================================
    /**
     * @internal
     * Init SubInstances like groups or selectors
     */
    initSubInstances(type) {
        const subInstance = internal_1.copy(this.config[type]) || {};
        let subInstanceObject = {};
        // If subInstance is array transform it to an object with the fitting class
        if (Array.isArray(subInstance)) {
            for (let i = 0; i < subInstance.length; i++) {
                let instance;
                switch (type) {
                    case "groups":
                        instance = new internal_1.Group(this.agileInstance(), this, [], { key: subInstance[i] });
                        break;
                    case "selectors":
                        instance = new internal_1.Selector(this, subInstance[i], { key: subInstance[i] });
                        break;
                    default:
                        instance = 'unknown';
                }
                subInstanceObject[subInstance[i]] = instance;
            }
        }
        else {
            // If subInstance is Object.. set subInstanceObject to subInstance
            subInstanceObject = subInstance;
        }
        // If groups.. add default group
        if (type === "groups") {
            if (!subInstanceObject[this.config.defaultGroupKey || 'default'])
                subInstanceObject[this.config.defaultGroupKey || 'default'] = new internal_1.Group(this.agileInstance(), this, [], { key: this.config.defaultGroupKey || 'default' });
        }
        const keys = Object.keys(subInstanceObject);
        for (let key of keys) {
            // Set key to property name if it isn't set yet
            if (!subInstanceObject[key].key)
                subInstanceObject[key].key = key;
        }
        // Set Collection instance
        this[type] = subInstanceObject;
    }
    //=========================================================================================================
    // Collect
    //=========================================================================================================
    /**
     * Collect iterable data into this collection.
     * Note: Data items must include a primary key (id)
     */
    collect(items, groups, options = {}) {
        const _items = internal_1.normalizeArray(items);
        const _groups = internal_1.normalizeArray(groups);
        const defaultGroupKey = this.config.defaultGroupKey || 'default';
        const groupsToRebuild = new Set();
        // Assign defaults to options
        options = internal_1.defineConfig(options, {
            method: 'push',
            background: false,
            patch: false
        });
        // Add default group if it hasn't been added (default group contains all items)
        if (_groups.findIndex(groupName => groupName === defaultGroupKey) === -1)
            _groups.push(defaultGroupKey);
        // Create not existing Groups
        _groups.forEach(groupName => !this.groups[groupName] && this.createGroup(groupName));
        _items.forEach((item, index) => {
            // Check if the item already exists in the Collection
            const itemExists = !!this.data[item[this.config.primaryKey || 'id']];
            // Save items into Collection
            let key = this.saveData(item, { patch: options.patch, background: options.background });
            // Return if key doesn't exist (something went wrong in saveData, Note: Error will be logged in saveData)
            if (!key)
                return;
            // Call forEachItem method
            if (options.forEachItem)
                options.forEachItem(item, key, index);
            // If item didn't exist.. check if the itemKey has already been added to a group before -> group need rebuild to has correct output
            if (!itemExists) {
                const groupKeys = Object.keys(this.groups);
                groupKeys.forEach(groupName => {
                    // Get Group
                    const group = this.getGroup(groupName);
                    // Check if itemKey exists in Group if so push it to groupsToRebuild
                    if (group.value.findIndex(primaryKey => primaryKey === item[this.config.primaryKey || 'id']) !== -1)
                        groupsToRebuild.add(group);
                });
            }
            // Add key to groups
            _groups.forEach(groupName => {
                // @ts-ignore
                this.groups[groupName].add(key, { method: options.method, background: options.background });
            });
        });
        // Rebuild groups
        groupsToRebuild.forEach(group => {
            // Rebuild Group
            group.build();
            // Force Rerender to get the correct output in components
            if (!options.background)
                group.ingest({ forceRerender: true });
        });
    }
    //=========================================================================================================
    // Update
    //=========================================================================================================
    /**
     * * Update data by updateKey(id) in a Agile Collection
     */
    update(updateKey, changes, options = {}) {
        // If item does not exist, return
        if (!this.data.hasOwnProperty(updateKey)) {
            console.error(`Agile: PrimaryKey '${updateKey} doesn't exist in collection `, this);
            return undefined;
        }
        // Assign defaults to config
        options = internal_1.defineConfig(options, {
            addNewProperties: false,
            background: false
        });
        const itemState = this.data[updateKey];
        const currentItemValue = internal_1.copy(itemState.value);
        const primaryKey = this.config.primaryKey || '';
        // Merge current Item value with changes
        const finalItemValue = internal_1.flatMerge(currentItemValue, changes, { addNewProperties: options.addNewProperties });
        // Check if something has changed (stringifying because of possible object or array)
        if (JSON.stringify(finalItemValue) === JSON.stringify(itemState.nextState))
            return this.data[finalItemValue[primaryKey]];
        // Assign finalItemStateValue to nextState
        itemState.nextState = finalItemValue;
        // Set State to nextState
        itemState.ingest({ background: options.background });
        // If data key changes update it properly
        if (currentItemValue[primaryKey] !== finalItemValue[primaryKey])
            this.updateItemPrimaryKeys(currentItemValue[primaryKey], finalItemValue[primaryKey], { background: options.background });
        // Rebuild all groups that includes the primaryKey
        this.rebuildGroupsThatIncludePrimaryKey(finalItemValue[primaryKey], { background: options.background });
        // Return data at primaryKey (updated State)
        return this.data[finalItemValue[primaryKey]];
    }
    //=========================================================================================================
    // Create Group
    //=========================================================================================================
    /**
     * Create a group instance on this collection
     */
    createGroup(groupName, initialItems) {
        // Check if Group already exist
        if (this.groups.hasOwnProperty(groupName)) {
            console.warn(`Agile: The Group with the name ${groupName} already exists!`);
            return this.groups[groupName];
        }
        // Create new Group
        const group = new internal_1.Group(this.agileInstance(), this, initialItems, { key: groupName });
        // Add new Group to groups
        this.groups[groupName] = group;
        // Log Job
        if (this.agileInstance().config.logJobs)
            console.log(`Agile: Created Group called '${groupName}'`, group);
        return group;
    }
    //=========================================================================================================
    // Create Selector
    //=========================================================================================================
    /**
     * Create a selector instance on this collection
     */
    createSelector(selectorName, id) {
        // Check if Selector already exist
        if (this.selectors.hasOwnProperty(selectorName)) {
            console.warn(`Agile: The Selector with the name ${selectorName} already exists!`);
            return this.selectors[selectorName];
        }
        // Create new Selector
        const selector = new internal_1.Selector(this, id, { key: selectorName });
        // Add new Selector to selectors
        this.selectors[selectorName] = selector;
        // Log Job
        if (this.agileInstance().config.logJobs)
            console.log(`Agile: Created Selector called '${selectorName}'`, selector);
        return selector;
    }
    //=========================================================================================================
    // Get Group
    //=========================================================================================================
    /**
     * Return an group from this collection as Group instance (extends State)
     */
    getGroup(groupName) {
        // Check if group exists
        if (this.groups[groupName])
            return this.groups[groupName];
        console.warn(`Agile: Group with name '${groupName}' doesn't exist!`);
        // Return empty group because it might get annoying to handle with undefined (can check if it exists with group.exists)
        const group = new internal_1.Group(this.agileInstance(), this, [], { key: 'dummy' });
        group.isPlaceholder = true;
        return group;
    }
    //=========================================================================================================
    // Get Selector
    //=========================================================================================================
    /**
     * Return an selector from this collection as Selector instance (extends State)
     */
    getSelector(selectorName) {
        // Check if selector exists
        if (this.selectors[selectorName])
            return this.selectors[selectorName];
        console.warn(`Agile: Selector with name '${selectorName}' doesn't exist!`);
        return undefined;
    }
    //=========================================================================================================
    // Remove
    //=========================================================================================================
    /**
     * Remove fromGroups or everywhere
     */
    remove(primaryKeys) {
        return {
            fromGroups: (groups) => this.removeFromGroups(primaryKeys, groups),
            everywhere: () => this.removeData(primaryKeys)
        };
    }
    //=========================================================================================================
    // Find By Id
    //=========================================================================================================
    /**
     * Return an item from this collection by primaryKey as Item instance (extends State)
     */
    findById(id) {
        if (!this.data.hasOwnProperty(id) || !this.data[id].exists)
            return undefined;
        // Add state to foundState (for auto tracking used states in computed functions)
        if (this.agileInstance().runtime.trackState)
            this.agileInstance().runtime.foundStates.add(this.data[id]);
        // Return data by id
        return this.data[id];
    }
    //=========================================================================================================
    // Get Value By Id
    //=========================================================================================================
    /**
     * Return a value from this collection by primaryKey
     */
    getValueById(id) {
        let data = this.findById(id);
        if (!data)
            return undefined;
        return data.value;
    }
    //=========================================================================================================
    // Persist
    //=========================================================================================================
    /**
     * Saves the collection in the local storage or in a own configured storage
     * @param key - the storage key (if no key passed it will take the collection key)
     */
    persist(key) {
        perstist_1.persistValue(this, key).then((value) => {
            this.isPersistCollection = value;
        });
        return this;
    }
    //=========================================================================================================
    // Group
    //=========================================================================================================
    /**
     * Create a group instance under this collection (can be used in function based config)
     */
    Group(initialItems, config) {
        return new internal_1.Group(this.agileInstance(), this, initialItems, config);
    }
    //=========================================================================================================
    // Selector
    //=========================================================================================================
    /**
     * Create a selector instance under this collection (can be used in function based config)
     */
    Selector(initialSelection, options) {
        return new internal_1.Selector(this, initialSelection, options);
    }
    //=========================================================================================================
    // Update Data Key
    //=========================================================================================================
    /**
     * @internal
     * This will properly change the key of a collection item
     */
    updateItemPrimaryKeys(oldKey, newKey, options) {
        // If oldKey and newKey are the same, return
        if (oldKey === newKey)
            return;
        // Assign defaults to config
        options = internal_1.defineConfig(options, {
            background: false
        });
        // Create copy of data
        const dataCopy = this.data[oldKey];
        // Delete old reference
        delete this.data[oldKey];
        // Apply the data into data with new key
        this.data[newKey] = dataCopy;
        // Update Groups
        for (let groupName in this.groups) {
            // Get Group
            const group = this.getGroup(groupName);
            // If Group does not contain oldKey, continue
            if (group.value.findIndex(key => key === oldKey) === -1)
                continue;
            // Replace the primaryKey at current index
            group.nextState.splice(group.nextState.indexOf(oldKey), 1, newKey);
            // Set State(Group) to nextState
            group.ingest({ background: options === null || options === void 0 ? void 0 : options.background });
        }
        // Update Selector
        for (let selectorName in this.selectors) {
            // Get Selector
            const selector = this.getSelector(selectorName);
            if (!selector)
                continue;
            // If Selector doesn't watch on the oldKey, continue
            if (selector.id !== oldKey)
                continue;
            // Replace the oldKey with the newKey
            selector.select(newKey, { background: options === null || options === void 0 ? void 0 : options.background });
        }
    }
    //=========================================================================================================
    // Remove From Groups
    //=========================================================================================================
    /**
     * @internal
     * Deletes Data from Groups
     */
    removeFromGroups(primaryKeys, groups) {
        const _primaryKeys = internal_1.normalizeArray(primaryKeys);
        const _groups = internal_1.normalizeArray(groups);
        _groups.forEach(groupKey => {
            // Return if group doesn't exist in collection
            if (!this.groups[groupKey]) {
                console.error(`Agile: Couldn't find group('${groupKey}) in collection`, this);
                return;
            }
            // Remove primaryKeys from Group
            _primaryKeys.forEach(primaryKey => {
                const group = this.getGroup(groupKey);
                group.remove(primaryKey);
            });
        });
    }
    //=========================================================================================================
    // Delete Data
    //=========================================================================================================
    /**
     * @internal
     * Deletes data directly form the collection
     */
    removeData(primaryKeys) {
        const _primaryKeys = internal_1.normalizeArray(primaryKeys);
        const groupKeys = Object.keys(this.groups);
        const selectorKeys = Object.keys(this.selectors);
        const itemKeys = Object.keys(this.data);
        _primaryKeys.forEach(itemKey => {
            // Check if primaryKey exists in collection, if not return
            if (itemKeys.findIndex(key => itemKey.toString() === key) === -1) {
                console.error(`Agile: Couldn't find primaryKey '${itemKey}' in collection`, this);
                return;
            }
            // Remove primaryKey from Groups (have to be above deleting the data because.. the remove function needs to know if the data exists or not)
            groupKeys.forEach(groupKey => {
                this.groups[groupKey].remove(itemKey);
            });
            // Remove Selectors with primaryKey
            selectorKeys.forEach(selectorKey => {
                delete this.selectors[selectorKey];
            });
            // Remove primaryKey from collection data
            delete this.data[itemKey];
            // Decrease size
            this.size--;
            // Storage
            perstist_1.removeItem(itemKey, this);
        });
    }
    //=========================================================================================================
    // Save Data
    //=========================================================================================================
    /**
     * @internal
     * Save data directly into the collection
     */
    saveData(data, options = {}) {
        // Transform data to any because otherwise I have many type errors (because not defined object)
        // https://stackoverflow.com/questions/57350092/string-cant-be-used-to-index-type
        const _data = data;
        // Assign defaults to options
        options = internal_1.defineConfig(options, {
            patch: false,
            background: false,
        });
        // Get primaryKey (default: 'id')
        const primaryKey = this.config.primaryKey || 'id';
        const itemKey = _data[primaryKey];
        // Check if data is object if not return
        if (!internal_1.isValidObject(_data)) {
            console.error("Agile: Collections items has to be an object for now!");
            return null;
        }
        // Check if data has primaryKey
        if (!_data.hasOwnProperty(primaryKey)) {
            console.error("Agile: Collections items need a own primaryKey. Here " + this.config.primaryKey);
            return null;
        }
        // Create reference of data at the data key
        let item = this.data[itemKey];
        // If the data already exists and config is to patch, patch data
        if (item && options.patch)
            item.patch(_data, { background: options.background });
        // If the data already exists and no config, overwrite data
        else if (item)
            item.set(_data, { background: options.background });
        // If data does not exist.. create new Item set and increase the size
        else {
            item = new internal_1.Item(this, _data);
            this.size++;
        }
        // Set item at data itemKey
        this.data[itemKey] = item;
        // Storage
        perstist_1.setItem(itemKey, this);
        return itemKey;
    }
    //=========================================================================================================
    // Rebuild Groups That Includes Primary Key
    //=========================================================================================================
    /**
     * @internal
     * Rebuild the Groups which contains the primaryKey
     */
    rebuildGroupsThatIncludePrimaryKey(primaryKey, options) {
        // Assign defaults to config
        options = internal_1.defineConfig(options, {
            background: false,
            forceRerender: !(options === null || options === void 0 ? void 0 : options.background) // forceRerender false.. because forceRerender has more weight than background in runtime
        });
        // Rebuild groups that includes primaryKey
        for (let groupKey in this.groups) {
            // Get Group
            const group = this.getGroup(groupKey);
            // Check if group contains primaryKey if so rebuild it
            if (group.has(primaryKey))
                group.ingest({ background: options === null || options === void 0 ? void 0 : options.background, forceRerender: options === null || options === void 0 ? void 0 : options.forceRerender });
        }
    }
}
exports.Collection = Collection;
