"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateGroup = exports.removeItem = exports.setItem = exports.persistValue = void 0;
const storageItemKeyTemplate = '_${collectionKey}_item_${itemKey}';
const storageGroupKeyTemplate = '_${collectionKey}_group_${groupKey}';
//=========================================================================================================
// Persist Value
//=========================================================================================================
/**
 * Will persist the 'collection' into the configured storage with the key or if no key passed the state key
 */
function persistValue(collection, key) {
    return __awaiter(this, void 0, void 0, function* () {
        // Validate Key
        const tempKey = validateKey(collection, key);
        if (!tempKey) {
            console.error("Agile: If your State has no key provided before using persist.. you have to provide a key here!");
            return false;
        }
        key = tempKey;
        // Get Storage
        const storage = collection.agileInstance().storage;
        // Add Collection to persistedCollections in Storage
        storage.persistedCollections.add(collection);
        // Call Handle which decides weather it has to add the storage value to the state or save the state into the storage
        yield handleStorageValue(storage, collection);
        return true;
    });
}
exports.persistValue = persistValue;
//=========================================================================================================
// Set/Update Item
//=========================================================================================================
/**
 * Set/Update Item in Storage
 */
function setItem(itemKey, collection) {
    return __awaiter(this, void 0, void 0, function* () {
        const storage = collection.agileInstance().storage;
        if (collection.isPersistCollection && collection.key) {
            // Create Item Storage Key
            const itemStorageKey = storageItemKeyTemplate
                .replace('${collectionKey}', (collection.key || 'unknown').toString())
                .replace('${itemKey}', itemKey.toString());
            // Get Value
            const item = collection.data[itemKey];
            if (!item)
                return;
            // Log Job
            if (collection.agileInstance().config.logJobs)
                console.log(`Agile Storage(Set Item): ${itemStorageKey}`, item.getPersistableValue());
            // Set StorageValue
            yield storage.set(itemStorageKey, item.getPersistableValue());
        }
    });
}
exports.setItem = setItem;
//=========================================================================================================
// Remove Item
//=========================================================================================================
/**
 * Removes Item from Storage
 */
function removeItem(itemKey, collection) {
    const storage = collection.agileInstance().storage;
    if (collection.isPersistCollection && collection.key) {
        // Create Item Storage Key
        const itemStorageKey = storageItemKeyTemplate
            .replace('${collectionKey}', (collection.key || 'unknown').toString())
            .replace('${itemKey}', itemKey.toString());
        // Log Job
        if (collection.agileInstance().config.logJobs)
            console.log(`Agile Storage(Remove Item): ${itemStorageKey}`);
        // Remove StorageValue
        storage.remove(itemStorageKey);
    }
}
exports.removeItem = removeItem;
//=========================================================================================================
// Update Group
//=========================================================================================================
/**
 * Updates the Group in the storage
 */
function updateGroup(groupKey, collection) {
    const storage = collection.agileInstance().storage;
    if (collection.isPersistCollection && collection.key) {
        // Create Group Storage Key
        const groupStorageKey = storageGroupKeyTemplate
            .replace('${collectionKey}', (collection.key || 'unknown').toString())
            .replace('${groupKey}', groupKey.toString());
        // Get group
        const group = collection.groups[groupKey];
        if (!group)
            return;
        // Log Job
        if (collection.agileInstance().config.logJobs)
            console.log(`Agile Storage(Update Group): ${groupStorageKey}`, group.value);
        // Set StorageValue
        storage.set(groupStorageKey, group.value);
    }
}
exports.updateGroup = updateGroup;
//=========================================================================================================
// Helper
//=========================================================================================================
function validateKey(collection, key) {
    // Get key from State key
    if (!key && collection.key)
        return collection.key;
    // Return null if no key can be found
    if (!key)
        return null;
    // Set this storage key as state key
    collection.key = key;
    return key;
}
function handleStorageValue(storage, collection) {
    return __awaiter(this, void 0, void 0, function* () {
        // Get Collection Data
        const collectionData = yield loadCollectionData(collection);
        // Log Job
        if (collection.agileInstance().config.logJobs)
            console.log("Agile Storage(CollectionData):", collectionData);
        // Load Collection Items
        yield loadCollectionItems(collectionData, collection);
        // Load Collection Groups
        yield loadCollectionGroups(collectionData, collection);
    });
}
function loadCollectionData(collection) {
    return __awaiter(this, void 0, void 0, function* () {
        const storage = collection.agileInstance().storage;
        const key = collection.key;
        const storageValue = yield storage.get(key);
        // If the collectionData doesn't exist in the storage yet.. create it
        if (!storageValue) {
            const collectionStorageData = {
                data: Object.keys(collection.data),
                groups: Object.keys(collection.groups)
            };
            storage.set(key, collectionStorageData);
            return collectionStorageData;
        }
        return storageValue;
    });
}
function loadCollectionItems(collectionStorageData, collection) {
    return __awaiter(this, void 0, void 0, function* () {
        const storage = collection.agileInstance().storage;
        for (let itemKey of collectionStorageData.data) {
            // Create Item Storage Key
            const itemStorageKey = storageItemKeyTemplate
                .replace('${collectionKey}', (collection.key || 'unknown').toString())
                .replace('${itemKey}', itemKey.toString());
            // Get StorageValue
            const storageValue = yield storage.get(itemStorageKey);
            // Log Job
            if (collection.agileInstance().config.logJobs)
                console.log(`Agile Storage: ${itemStorageKey}`, storageValue);
            // If the value doesn't exist in the storage yet.. create it
            if (!storageValue) {
                storage.set(itemStorageKey, collection.data[itemKey].getPersistableValue());
                return;
            }
            // If the value already exists in the storage.. load it into the collection
            collection.collect(storageValue);
        }
    });
}
function loadCollectionGroups(collectionStorageData, collection) {
    return __awaiter(this, void 0, void 0, function* () {
        const storage = collection.agileInstance().storage;
        for (let groupKey of collectionStorageData.groups) {
            // Create Group Storage Key
            const groupStorageKey = storageGroupKeyTemplate
                .replace('${collectionKey}', (collection.key || 'unknown').toString())
                .replace('${groupKey}', groupKey.toString());
            // Get StorageValue
            const storageValue = yield storage.get(groupStorageKey);
            // Log Job
            if (collection.agileInstance().config.logJobs)
                console.log(`Agile Storage: ${groupStorageKey}`, storageValue);
            // If the value doesn't exist in the storage yet.. create it
            if (!storageValue) {
                storage.set(groupStorageKey, collection.groups[groupKey].value);
                return;
            }
            // - If the value already exists in the storage.. load it into the collection
            // Get Group
            const group = collection.groups[groupKey];
            // If group doesn't exists, Create Group
            if (!group) {
                collection.createGroup(groupKey, storageValue);
                continue;
            }
            // If group exists, add items into group
            for (let key of storageValue) {
                if (!group.has(key)) {
                    group.add(key);
                }
            }
        }
    });
}
