"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.State = void 0;
const internal_1 = require("../internal");
const persist_1 = require("./persist");
class State {
    constructor(agileInstance, initialState, key, deps = []) {
        this.watchers = {};
        this.isSet = false; // Has been changed from initial value
        this.isPlaceholder = false; // Defines if the state is a placeholder or not
        this.agileInstance = () => agileInstance;
        this.initialState = initialState;
        this.dep = new internal_1.Dep(deps);
        this._key = key;
        this._value = initialState;
        this.previousState = initialState;
        this.nextState = initialState;
        this.persistSettings = {
            isPersisted: false
        };
    }
    set value(value) {
        this.set(value);
    }
    get value() {
        // Add state to foundState (for auto tracking used states in computed functions)
        if (this.agileInstance().runtime.trackState)
            this.agileInstance().runtime.foundStates.add(this);
        return this._value;
    }
    set key(value) {
        this._key = value;
    }
    get key() {
        return this._key;
    }
    //=========================================================================================================
    // Set
    //=========================================================================================================
    /**
     * Directly set state to a new value
     */
    set(value, options = {}) {
        // Assign defaults to options
        options = internal_1.defineConfig(options, {
            sideEffects: true,
            background: false
        });
        // Check if Type is Correct
        if (this.valueType && !this.isCorrectType(value)) {
            console.warn(`Agile: Incorrect type (${typeof value}) was provided. Type fixed to ${this.valueType}`);
            return this;
        }
        // Check if something has changed (stringifying because of possible object or array)
        if (JSON.stringify(this.value) === JSON.stringify(value))
            return this;
        // Ingest updated value
        this.agileInstance().runtime.ingest(this, value, {
            background: options.background,
            sideEffects: options.sideEffects
        });
        return this;
    }
    //=========================================================================================================
    // Ingest
    //=========================================================================================================
    /**
     * @internal
     * Will ingest the nextState or the computedValue (rebuilds state)
     */
    ingest(options = {}) {
        // Assign defaults to options
        options = internal_1.defineConfig(options, {
            sideEffects: true,
            background: false,
            forceRerender: false
        });
        this.agileInstance().runtime.ingest(this, this.agileInstance().runtime.internalIngestKey, {
            background: options.background,
            sideEffects: options.sideEffects,
            forceRerender: options.forceRerender
        });
    }
    //=========================================================================================================
    // Type
    //=========================================================================================================
    /**
     * This is thought for js users.. because ts users can set the type in <>
     * @param type - wished type of the state
     */
    type(type) {
        // Supported types
        const supportedTypes = ['String', 'Boolean', 'Array', 'Object', 'Number'];
        // Check if type is a supported Type
        if (supportedTypes.findIndex(supportedType => supportedType === type.name) === -1) {
            console.warn(`Agile: '${type}' is not supported! Supported types: String, Boolean, Array, Object, Number`);
            return this;
        }
        // Set valueType
        this.valueType = type.name.toLowerCase();
        return this;
    }
    //=========================================================================================================
    // Undo
    //=========================================================================================================
    /**
     * Will set the state to the previous State
     */
    undo() {
        this.set(this.previousState);
    }
    //=========================================================================================================
    // Reset
    //=========================================================================================================
    /**
     * Will reset the state to the initial value
     */
    reset() {
        // Remove State from Storage (because it is than the initial State again and there is no need to save it anymore)
        if (this.persistSettings.isPersisted && this.persistSettings.persistKey)
            this.agileInstance().storage.remove(this.persistSettings.persistKey);
        // Set State to initial State
        this.set(this.initialState);
        return this;
    }
    //=========================================================================================================
    // Patch
    //=========================================================================================================
    /**
     * Will merge the changes into the state
     */
    patch(targetWithChanges, options = {}) {
        // Check if state is object.. because only objects can use the patch method
        if (!internal_1.isValidObject(this.nextState)) {
            console.warn("Agile: You can't use the patch method on a non object state!");
            return this;
        }
        // Check if targetWithChanges is an Object.. because you can only patch objects into the State Object
        if (!internal_1.isValidObject(targetWithChanges)) {
            console.warn("Agile: TargetWithChanges has to be an object!");
            return this;
        }
        // Assign defaults to options
        options = internal_1.defineConfig(options, {
            addNewProperties: true,
            background: false
        });
        // Merge targetWithChanges into next State
        this.nextState = internal_1.flatMerge(this.nextState, targetWithChanges, options);
        // Check if something has changed (stringifying because of possible object or array)
        if (JSON.stringify(this.value) === JSON.stringify(this.nextState))
            return this;
        // Set State to nextState
        this.ingest({ background: options.background });
        this.isSet = this.nextState !== this.initialState;
        return this;
    }
    //=========================================================================================================
    // Watch
    //=========================================================================================================
    /**
     * Will always be called if the state changes
     * @param key - The key of the watch method
     * @param callback - The callback function
     */
    watch(key, callback) {
        // Check if callback is a function  (js)
        if (typeof callback !== 'function') {
            console.error('Agile: A watcher callback function has to be an function!');
            return this;
        }
        // Add callback with key to watchers
        this.watchers[key] = callback;
        return this;
    }
    //=========================================================================================================
    // Remove Watcher
    //=========================================================================================================
    /**
     * Removes a watcher called after the key
     * @param key - the key of the watcher function
     */
    removeWatcher(key) {
        delete this.watchers[key];
        return this;
    }
    //=========================================================================================================
    // Persist
    //=========================================================================================================
    /**
     * Saves the state in the local storage or in a own configured storage
     * @param key - the storage key (if no key passed it will take the state key)
     */
    persist(key) {
        persist_1.persistValue(this, key);
        return this;
    }
    //=========================================================================================================
    // Copy
    //=========================================================================================================
    /**
     * Returns a fresh copy of the current value
     */
    copy() {
        return internal_1.copy(this.value);
    }
    //=========================================================================================================
    // Exists
    //=========================================================================================================
    /**
     * Checks if the State exists
     */
    get exists() {
        // Check if the value is not undefined and that the state is no placeholder
        return this.getPublicValue() !== undefined && !this.isPlaceholder;
    }
    //=========================================================================================================
    // Get Public Value
    //=========================================================================================================
    /**
     * @internal
     *  Returns 100% the public value of a state because at some points (group) the _value contains only keys
     */
    getPublicValue() {
        if (this.output !== undefined)
            return this.output;
        return this._value;
    }
    //=========================================================================================================
    // Private Write
    //=========================================================================================================
    /**
     * @internal
     *  Will set a new _value and handles the stuff around like storage, ..
     */
    privateWrite(value) {
        this._value = internal_1.copy(value);
        this.nextState = internal_1.copy(value);
        // Save changes in Storage
        persist_1.updateValue(this);
    }
    //=========================================================================================================
    // Get Persistable Value
    //=========================================================================================================
    /**
     * @internal
     *  Will return the perstiable Value of this state..
     *  some classes which extends state might have another peristiableValue than this.value (like the selector)
     */
    getPersistableValue() {
        return this.value;
    }
    //=========================================================================================================
    // Is Correct Type
    //=========================================================================================================
    /**
     * @internal
     * Checks if the 'value' has the same type as state.value
     */
    isCorrectType(value) {
        let type = typeof value;
        return type === this.valueType;
    }
}
exports.State = State;
