"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Storage = void 0;
const internal_1 = require("../internal");
class Storage {
    constructor(agileInstance, storageConfig) {
        this.isAsync = false;
        this.storageReady = false;
        this.storageType = 'localStorage';
        this.storagePrefix = 'agile';
        this.persistedStates = new Set();
        this.persistedCollections = new Set();
        this.agileInstance = () => agileInstance;
        this.storageConfig = storageConfig;
        // Set custom Storage prefix
        if (storageConfig.prefix)
            this.storagePrefix = storageConfig.prefix;
        // Set custom Storage functions
        if (storageConfig.methods)
            this.storageType = 'custom';
        if (storageConfig.async)
            this.isAsync = true;
        // Instantiate Custom Storage
        if (this.storageType === 'custom')
            this.instantiateCustomStorage();
        // Instantiate Local Storage
        if (this.storageType === 'localStorage')
            this.instantiateLocalStorage();
    }
    //=========================================================================================================
    // Instantiate Local Storage
    //=========================================================================================================
    /**
     * This instantiate the Local Storage
     */
    instantiateLocalStorage() {
        // Check if Local Storage is Available (For instance in ReactNative it doesn't exist)
        if (!this.localStorageAvailable()) {
            console.warn("Agile: Local Storage is here not available.. to use the Storage functionality please provide a custom Storage!");
            return;
        }
        // Set StorageMethods to LocalStorageMethods
        this.storageConfig.methods = {
            get: localStorage.getItem.bind(localStorage),
            set: localStorage.setItem.bind(localStorage),
            remove: localStorage.removeItem.bind(localStorage)
        };
        this.storageReady = true;
    }
    //=========================================================================================================
    // Instantiate Custom Storage
    //=========================================================================================================
    /**
     * This instantiate the Custom Storage
     */
    instantiateCustomStorage() {
        var _a, _b, _c, _d, _e, _f;
        // Check Get Function
        if (!internal_1.isFunction((_a = this.storageConfig.methods) === null || _a === void 0 ? void 0 : _a.get)) {
            console.error("Agile: Your GET StorageMethod isn't valid!");
            return;
        }
        // Check Set Function
        if (!internal_1.isFunction((_b = this.storageConfig.methods) === null || _b === void 0 ? void 0 : _b.set)) {
            console.error("Agile: Your SET StorageMethod isn't valid!");
            return;
        }
        // Check Remove Function
        if (!internal_1.isFunction((_c = this.storageConfig.methods) === null || _c === void 0 ? void 0 : _c.remove)) {
            console.error("Agile: Your REMOVE StorageMethod isn't valid!");
            return;
        }
        // Check if one function is async if so set is Async to true
        if (internal_1.isAsyncFunction((_d = this.storageConfig.methods) === null || _d === void 0 ? void 0 : _d.get) || internal_1.isAsyncFunction((_e = this.storageConfig.methods) === null || _e === void 0 ? void 0 : _e.set) || internal_1.isAsyncFunction((_f = this.storageConfig.methods) === null || _f === void 0 ? void 0 : _f.remove))
            this.isAsync = true;
        this.storageReady = true;
    }
    //=========================================================================================================
    // Get
    //=========================================================================================================
    /**
     * Gets the value provided by the key from the storage
     */
    get(key) {
        var _a;
        if (!this.storageReady || !((_a = this.storageConfig.methods) === null || _a === void 0 ? void 0 : _a.get))
            return;
        // Async Get
        if (this.isAsync)
            return new Promise((resolve, reject) => {
                var _a;
                (_a = this.storageConfig.methods) === null || _a === void 0 ? void 0 : _a.get(this.getStorageKey(key)).then((res) => {
                    // If result is no Json
                    if (!internal_1.isJsonString(res))
                        return resolve(res);
                    // Format Json to Object
                    resolve(JSON.parse(res));
                }).catch(reject);
            });
        // Normal Get
        const res = this.storageConfig.methods.get(this.getStorageKey(key));
        if (internal_1.isJsonString(res))
            return JSON.parse(res);
        return res;
    }
    //=========================================================================================================
    // Set
    //=========================================================================================================
    /**
     * Sets the value into the storage
     */
    set(key, value) {
        var _a;
        if (!this.storageReady || !((_a = this.storageConfig.methods) === null || _a === void 0 ? void 0 : _a.set))
            return;
        this.storageConfig.methods.set(this.getStorageKey(key), JSON.stringify(value));
    }
    //=========================================================================================================
    // Remove
    //=========================================================================================================
    /**
     * Deletes the value that is stored with the key
     */
    remove(key) {
        var _a;
        if (!this.storageReady || !((_a = this.storageConfig.methods) === null || _a === void 0 ? void 0 : _a.remove))
            return;
        this.storageConfig.methods.remove(this.getStorageKey(key));
    }
    //=========================================================================================================
    // Helper
    //=========================================================================================================
    getStorageKey(key) {
        return `_${this.storagePrefix}_${key}`;
    }
    localStorageAvailable() {
        try {
            localStorage.setItem('_', '_');
            localStorage.removeItem('_');
            return true;
        }
        catch (e) {
            return false;
        }
    }
}
exports.Storage = Storage;
