"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.flatMerge = exports.defineConfig = exports.isJsonString = exports.isValidUrl = exports.isAsyncFunction = exports.isFunction = exports.getAgileInstance = exports.normalizeArray = exports.isValidObject = exports.copy = void 0;
const internal_1 = require("./internal");
function copy(value) {
    if (Array.isArray(value))
        return [...value];
    if (isValidObject(value))
        return Object.assign({}, value);
    return value;
}
exports.copy = copy;
//=========================================================================================================
// Is Valid Object
//=========================================================================================================
/**
 * Checks if an Object is an valid object for Agile
 * https://stackoverflow.com/questions/12996871/why-does-typeof-array-with-objects-return-object-and-not-array
 */
function isValidObject(value) {
    function isHTMLElement(obj) {
        try {
            return obj instanceof HTMLElement;
        }
        catch (e) {
            return typeof obj === 'object' && obj.nodeType === 1 && typeof obj.style === 'object' && typeof obj.ownerDocument === 'object';
        }
    }
    return value !== null && typeof value === 'object' && !isHTMLElement(value) && !Array.isArray(value);
}
exports.isValidObject = isValidObject;
//=========================================================================================================
// Normalize Array
//=========================================================================================================
/**
 * Convert item into an array
 */
function normalizeArray(items) {
    // Return empty array if no items
    if (!items)
        return [];
    return Array.isArray(items) ? items : [items];
}
exports.normalizeArray = normalizeArray;
//=========================================================================================================
// Get Instance
//=========================================================================================================
/**
 * Tries to get AgileInstance from instance(State, Collection)
 */
function getAgileInstance(instance) {
    try {
        // Return state agileInstance
        if (instance instanceof internal_1.State)
            return instance.agileInstance();
        if (instance instanceof internal_1.Event)
            return instance.agileInstance();
        if (instance instanceof internal_1.Collection)
            return instance.agileInstance();
        // Return the globalBind agile instance
        // @ts-ignore
        return globalThis.__agile;
    }
    catch (e) {
        // fail silently
    }
    return null;
}
exports.getAgileInstance = getAgileInstance;
//=========================================================================================================
// Is Function
//=========================================================================================================
/**
 * Checks if func is a function
 */
function isFunction(func) {
    return typeof func === 'function';
}
exports.isFunction = isFunction;
//=========================================================================================================
// Is Async Function
//=========================================================================================================
/**
 * Checks if func is a async function
 */
function isAsyncFunction(func) {
    return isFunction(func) && func.constructor.name === 'AsyncFunction';
}
exports.isAsyncFunction = isAsyncFunction;
//=========================================================================================================
// Is Valid Url
// https://stackoverflow.com/questions/5717093/check-if-a-javascript-string-is-a-url
//=========================================================================================================
/**
 * Checks if url is valid
 */
function isValidUrl(url) {
    const pattern = new RegExp('^(https?:\\/\\/)?' + // protocol
        '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|' + // domain name
        '((\\d{1,3}\\.){3}\\d{1,3}))' + // OR ip (v4) address
        '(\\:\\d+)?(\\/[-a-z\\d%_.~+]*)*' + // port and path
        '(\\?[;&a-z\\d%_.~+=-]*)?' + // query string
        '(\\#[-a-z\\d_]*)?$', 'i'); // fragment locator
    return pattern.test(url);
}
exports.isValidUrl = isValidUrl;
//=========================================================================================================
// Is Json String
//=========================================================================================================
/**
 * Checks if value is a valid JsonString
 */
function isJsonString(value) {
    try {
        JSON.parse(value);
    }
    catch (e) {
        return false;
    }
    return true;
}
exports.isJsonString = isJsonString;
//=========================================================================================================
// Define Config
//=========================================================================================================
/**
 * Will create a config (config) and merges default values (default) into this config (config)
 */
function defineConfig(config, defaults) {
    return Object.assign(Object.assign({}, defaults), config);
}
exports.defineConfig = defineConfig;
//=========================================================================================================
// Flat Merge
//=========================================================================================================
/**
 * Merged the items flat into the object
 */
function flatMerge(source, changes, config = {}) {
    // Copy Source to avoid reference
    const _source = copy(source);
    // Loop through changes object and merge changes into source
    let keys = Object.keys(changes);
    keys.forEach(property => {
        // @ts-ignore https://stackoverflow.com/questions/18452920/continue-in-cursor-foreach
        if (!config.addNewProperties && !_source[property])
            return;
        // @ts-ignore
        _source[property] = changes[property];
    });
    return _source;
}
exports.flatMerge = flatMerge;
